@@ -26,9 +26,9 @@ const DragMotionResult = {
 };
 
 const DRAG_CURSOR_MAP = {
-    0: Shell.Cursor.UNSUPPORTED_TARGET,
-    1: Shell.Cursor.COPY,
-    2: Shell.Cursor.MOVE
+    0: Shell.Cursor.DND_UNSUPPORTED_TARGET,
+    1: Shell.Cursor.DND_COPY,
+    2: Shell.Cursor.DND_MOVE
 };
 
 const DragDropResult = {
@@ -87,6 +87,10 @@ _Draggable.prototype = {
 
         this.actor.connect('destroy', Lang.bind(this, function() {
             this._actorDestroyed = true;
+            // If the drag actor is destroyed and we were going to fix
+            // up its hover state, fix up the parent hover state instead
+            if (this.actor == this._firstLeaveActor)
+                this._firstLeaveActor = this._dragOrigParent;
             if (this._dragInProgress)
                 this._cancelDrag(global.get_current_time());
             this.disconnectAll();
@@ -100,6 +104,14 @@ _Draggable.prototype = {
         this._buttonDown = false; // The mouse button has been pressed and has not yet been released.
         this._dragInProgress = false; // The drag has been started, and has not been dropped or cancelled yet.
         this._animationInProgress = false; // The drag is over and the item is in the process of animating to its original position (snapping back or reverting).
+
+        // During the drag, we eat enter/leave events so that actors don't prelight or show
+        // tooltips. But we remember the actors that we first left/last entered so we can
+        // fix up the hover state after the drag ends.
+        this._firstLeaveActor = null;
+        this._lastEnterActor = null;
+
+        this._eventsGrabbed = false;
     },
 
     _onButtonPress : function (actor, event) {
@@ -110,11 +122,11 @@ _Draggable.prototype = {
             return false;
 
         this._buttonDown = true;
-        // special case St.Clickable: grabbing the pointer would mess up the
+        // special case St.Button: grabbing the pointer would mess up the
         // internal state, so we start the drag manually on hover change
-        if (this.actor instanceof St.Clickable)
+        if (this.actor instanceof St.Button)
             this.actor.connect('notify::hover',
-                               Lang.bind(this, this._onClickableHoverChanged));
+                               Lang.bind(this, this._onButtonHoverChanged));
         else
             this._grabActor();
 
@@ -125,8 +137,8 @@ _Draggable.prototype = {
         return false;
     },
 
-    _onClickableHoverChanged: function(button) {
-        if (button.hover || !button.held)
+    _onButtonHoverChanged: function(button) {
+        if (button.hover || !button.pressed)
             return;
 
         button.fake_release();
@@ -142,18 +154,26 @@ _Draggable.prototype = {
 
     _ungrabActor: function() {
         Clutter.ungrab_pointer();
+        if (!this._onEventId)
+            return;
         this.actor.disconnect(this._onEventId);
         this._onEventId = null;
     },
 
     _grabEvents: function() {
-        Clutter.grab_pointer(_getEventHandlerActor());
-        Clutter.grab_keyboard(_getEventHandlerActor());
+        if (!this._eventsGrabbed) {
+            Clutter.grab_pointer(_getEventHandlerActor());
+            Clutter.grab_keyboard(_getEventHandlerActor());
+            this._eventsGrabbed = true;
+        }
     },
 
     _ungrabEvents: function() {
-        Clutter.ungrab_pointer();
-        Clutter.ungrab_keyboard();
+        if (this._eventsGrabbed) {
+            Clutter.ungrab_pointer();
+            Clutter.ungrab_keyboard();
+            this._eventsGrabbed = false;
+        }
     },
 
     _onEvent: function(actor, event) {
@@ -190,6 +210,11 @@ _Draggable.prototype = {
                 this._cancelDrag(event.get_time());
                 return true;
             }
+        } else if (event.type() == Clutter.EventType.LEAVE) {
+            if (this._firstLeaveActor == null)
+                this._firstLeaveActor = event.get_source();
+        } else if (event.type() == Clutter.EventType.ENTER) {
+            this._lastEnterActor = event.get_source();
         }
 
         return false;
@@ -213,7 +238,7 @@ _Draggable.prototype = {
         if (this._onEventId)
             this._ungrabActor();
         this._grabEvents();
-        global.set_cursor(Shell.Cursor.IN_DRAG);
+        global.set_cursor(Shell.Cursor.DND_IN_DRAG);
 
         this._dragX = this._dragStartX = stageX;
         this._dragY = this._dragStartY = stageY;
@@ -267,6 +292,7 @@ _Draggable.prototype = {
 
         this._dragActor.reparent(this.actor.get_stage());
         this._dragActor.raise_top();
+        Shell.util_set_hidden_from_pick(this._dragActor, true);
 
         this._dragOrigOpacity = this._dragActor.opacity;
         if (this._dragActorOpacity != undefined)
@@ -332,12 +358,8 @@ _Draggable.prototype = {
             this._dragActor.set_position(stageX + this._dragOffsetX,
                                          stageY + this._dragOffsetY);
 
-            // Because we want to find out what other actor is located at the current position of this._dragActor,
-            // we have to temporarily hide this._dragActor.
-            this._dragActor.hide();
             let target = this._dragActor.get_stage().get_actor_at_pos(Clutter.PickMode.ALL,
                                                                       stageX, stageY);
-            this._dragActor.show();
 
             // We call observers only once per motion with the innermost
             // target actor. If necessary, the observer can walk the
@@ -377,20 +399,16 @@ _Draggable.prototype = {
                 }
                 target = target.get_parent();
             }
-            global.set_cursor(Shell.Cursor.IN_DRAG);
+            global.set_cursor(Shell.Cursor.DND_IN_DRAG);
         }
 
         return true;
     },
 
     _dragActorDropped: function(event) {
-        // Find a drop target. Because we want to find out what other actor is located at
-        // the current position of this._dragActor, we have to temporarily hide this._dragActor.
-        this._dragActor.hide();
         let [dropX, dropY] = event.get_coords();
         let target = this._dragActor.get_stage().get_actor_at_pos(Clutter.PickMode.ALL,
                                                                   dropX, dropY);
-        this._dragActor.show();
 
         // We call observers only once per motion with the innermost
         // target actor. If necessary, the observer can walk the
@@ -447,24 +465,44 @@ _Draggable.prototype = {
         return true;
     },
 
-    // Get position of the drag actor's source if the source is still around,
-    // or return the original location if the actor itself was being dragged
-    // or the source is no longer around.
     _getRestoreLocation: function() {
-        let locX = this._snapBackX;
-        let locY = this._snapBackY;
+        let x, y, scale;
 
-        if (this._dragActorSource && this._dragActorSource.visible)
-            [locX, locY] = this._dragActorSource.get_transformed_position();
-        return [locX, locY];
+        if (this._dragActorSource && this._dragActorSource.visible) {
+            // Snap the clone back to its source
+            [x, y] = this._dragActorSource.get_transformed_position();
+            let [sourceScaledWidth, sourceScaledHeight] = this._dragActorSource.get_transformed_size();
+            scale = this._dragActor.width / sourceScaledWidth;
+        } else if (this._dragOrigParent) {
+            // Snap the actor back to its original position within
+            // its parent, adjusting for the fact that the parent
+            // may have been moved or scaled
+            let [parentX, parentY] = this._dragOrigParent.get_transformed_position();
+            let [parentWidth, parentHeight] = this._dragOrigParent.get_size();
+            let [parentScaledWidth, parentScaledHeight] = this._dragOrigParent.get_transformed_size();
+            let parentScale = parentScaledWidth / parentWidth;
+
+            x = parentX + parentScale * this._dragOrigX;
+            y = parentY + parentScale * this._dragOrigY;
+            scale = this._dragOrigScale * parentScale;
+        } else {
+            // Snap back actor to its original stage position
+            x = this._snapBackX;
+            y = this._snapBackY;
+            scale = this._snapBackScale;
+        }
+
+        return [x, y, scale];
     },
 
     _cancelDrag: function(eventTime) {
         this._dragInProgress = false;
-        let [snapBackX, snapBackY] = this._getRestoreLocation();
+        let [snapBackX, snapBackY, snapBackScale] = this._getRestoreLocation();
 
         if (this._actorDestroyed) {
             global.unset_cursor();
+            if (!this._buttonDown)
+                this._dragComplete();
             this.emit('drag-end', eventTime, false);
             return;
         }
@@ -474,8 +512,8 @@ _Draggable.prototype = {
         Tweener.addTween(this._dragActor,
                          { x: snapBackX,
                            y: snapBackY,
-                           scale_x: this._snapBackScale,
-                           scale_y: this._snapBackScale,
+                           scale_x: snapBackScale,
+                           scale_y: snapBackScale,
                            opacity: this._dragOrigOpacity,
                            time: SNAP_BACK_ANIMATION_TIME,
                            transition: 'easeOutQuad',
@@ -487,11 +525,11 @@ _Draggable.prototype = {
 
     _restoreDragActor: function(eventTime) {
         this._dragInProgress = false;
-        [restoreX, restoreY] = this._getRestoreLocation();
+        [restoreX, restoreY, restoreScale] = this._getRestoreLocation();
 
         // fade the actor back in at its original location
         this._dragActor.set_position(restoreX, restoreY);
-        this._dragActor.set_scale(this._snapBackScale, this._snapBackScale);
+        this._dragActor.set_scale(restoreScale, restoreScale);
         this._dragActor.opacity = 0;
 
         this._animationInProgress = true;
@@ -521,10 +559,36 @@ _Draggable.prototype = {
             this._dragComplete();
     },
 
+    // Actor is an actor we have entered or left during the drag; call
+    // st_widget_sync_hover on all StWidget ancestors
+    _syncHover: function(actor) {
+        while (actor) {
+            let parent = actor.get_parent();
+            if (actor instanceof St.Widget)
+                actor.sync_hover();
+
+            actor = parent;
+        }
+    },
+
     _dragComplete: function() {
+        if (!this._actorDestroyed)
+            Shell.util_set_hidden_from_pick(this._dragActor, false);
+
+        this._ungrabEvents();
+
+        if (this._firstLeaveActor) {
+            this._syncHover(this._firstLeaveActor);
+            this._firstLeaveActor = null;
+        }
+
+        if (this._lastEnterActor) {
+            this._syncHover(this._lastEnterActor);
+            this._lastEnterActor = null;
+        }
+
         this._dragActor = undefined;
         currentDraggable = null;
-        this._ungrabEvents();
     }
 };
 
