@@ -9,9 +9,11 @@ const St = imports.gi.St;
 const Gettext = imports.gettext.domain('gnome-shell');
 const _ = Gettext.gettext;
 
+const Config = imports.misc.config;
 const Main = imports.ui.main;
 const MessageTray = imports.ui.messageTray;
 const Params = imports.misc.params;
+const Util = imports.misc.util;
 
 let nextNotificationId = 1;
 
@@ -91,16 +93,6 @@ NotificationDaemon.prototype = {
     _init: function() {
         DBus.session.exportObject('/org/freedesktop/Notifications', this);
 
-        this._everAcquiredName = false;
-        DBus.session.acquire_name('org.freedesktop.Notifications',
-                                  // We pass MANY_INSTANCES so that if
-                                  // notification-daemon is running, we'll
-                                  // get queued behind it and then get the
-                                  // name after killing it below
-                                  DBus.MANY_INSTANCES,
-                                  Lang.bind(this, this._acquiredName),
-                                  Lang.bind(this, this._lostName));
-
         this._sources = {};
         this._senderToPid = {};
         this._notifications = {};
@@ -115,31 +107,6 @@ NotificationDaemon.prototype = {
             Lang.bind(this, this._onFocusAppChanged));
     },
 
-    _acquiredName: function() {
-        this._everAcquiredName = true;
-    },
-
-    _lostName: function() {
-        if (this._everAcquiredName)
-            log('Lost name org.freedesktop.Notifications!');
-        else if (GLib.getenv('GNOME_SHELL_NO_REPLACE'))
-            log('Failed to acquire org.freedesktop.Notifications');
-        else {
-            log('Failed to acquire org.freedesktop.Notifications; trying again');
-
-            // kill the notification-daemon. pkill is more portable
-            // than killall, but on Linux at least it won't match if
-            // you pass more than 15 characters of the process name...
-            // However, if you use the '-f' flag to match the entire
-            // command line, it will work, but we have to be careful
-            // in that case that we don't match 'gedit
-            // notification-daemon.c' or whatever...
-            let p = new Shell.Process({ args: ['pkill', '-f',
-                                               '^([^ ]*/)?(notification-daemon|notify-osd)$']});
-            p.run();
-        }
-    },
-
     _iconForNotificationData: function(icon, hints, size) {
         let textureCache = St.TextureCache.get_default();
 
@@ -153,9 +120,9 @@ NotificationDaemon.prototype = {
                 return new St.Icon({ icon_name: icon,
                                      icon_type: St.IconType.FULLCOLOR,
                                      icon_size: size });
-        } else if (hints.icon_data) {
+        } else if (hints['image-data']) {
             let [width, height, rowStride, hasAlpha,
-                 bitsPerSample, nChannels, data] = hints.icon_data;
+                 bitsPerSample, nChannels, data] = hints['image-data'];
             return textureCache.load_from_raw(data, data.length, hasAlpha,
                                               width, height, rowStride, size);
         } else {
@@ -245,6 +212,15 @@ NotificationDaemon.prototype = {
 
         hints = Params.parse(hints, { urgency: Urgency.NORMAL }, true);
 
+        // Be compatible with the various hints for image data
+        // 'image-data' is the latest name of this hint, introduced in 1.2
+        if (!hints['image-data']) {
+            if (hints['image_data'])
+                hints['image-data'] = hints['image_data']; // version 1.1 of the spec
+            else if (hints['icon_data'])
+                hints['image-data'] = hints['icon_data']; // previous versions of the spec
+        }
+
         let ndata = { appName: appName,
                       icon: icon,
                       summary: summary,
@@ -320,13 +296,22 @@ NotificationDaemon.prototype = {
                                                         { icon: iconActor,
                                                           bannerMarkup: true });
             ndata.notification = notification;
-            notification.connect('clicked', Lang.bind(this,
-                function(n) {
-                    this._emitNotificationClosed(id, NotificationClosedReason.DISMISSED);
-                }));
             notification.connect('destroy', Lang.bind(this,
-                function(n) {
+                function(n, reason) {
                     delete this._notifications[id];
+                    let notificationClosedReason;
+                    switch (reason) {
+                        case MessageTray.NotificationDestroyedReason.EXPIRED:
+                            notificationClosedReason = NotificationClosedReason.EXPIRED;
+                            break;
+                        case MessageTray.NotificationDestroyedReason.DISMISSED:
+                            notificationClosedReason = NotificationClosedReason.DISMISSED;
+                            break;
+                        case MessageTray.NotificationDestroyedReason.SOURCE_CLOSED:
+                            notificationClosedReason = NotificationClosedReason.APP_CLOSED;
+                            break;
+                    }
+                    this._emitNotificationClosed(id, notificationClosedReason);
                 }));
             notification.connect('action-invoked', Lang.bind(this,
                 function(n, actionId) {
@@ -367,10 +352,9 @@ NotificationDaemon.prototype = {
         let ndata = this._notifications[id];
         if (ndata) {
             if (ndata.notification)
-                ndata.notification.destroy();
+                ndata.notification.destroy(MessageTray.NotificationDestroyedReason.SOURCE_CLOSED);
             delete this._notifications[id];
         }
-        this._emitNotificationClosed(id, NotificationClosedReason.APP_CLOSED);
     },
 
     GetCapabilities: function() {
@@ -390,10 +374,10 @@ NotificationDaemon.prototype = {
 
     GetServerInformation: function() {
         return [
-            'GNOME Shell',
+            Config.PACKAGE_NAME,
             'GNOME',
-            '0.1', // FIXME, get this from somewhere
-            '1.0'
+            Config.PACKAGE_VERSION,
+            '1.2'
         ];
     },
 
@@ -451,6 +435,7 @@ Source.prototype = {
         MessageTray.Source.prototype._init.call(this, title);
 
         this._pid = pid;
+        this._appStateChangedId = 0;
         this._setApp();
         if (this.app)
             this.title = this.app.get_name();
@@ -475,6 +460,10 @@ Source.prototype = {
         if (!this.app)
             return;
 
+        // We only update the app if this.app is null, so we can't disconnect the old this._appStateChangedId
+        // even if it were non-zero for some reason.
+        this._appStateChangedId = this.app.connect('notify::state', Lang.bind(this,  this._appStateChanged));
+
         // Only override the icon if we were previously using
         // notification-based icons (ie, not a trayicon) or if it was unset before
         if (!this._isTrayIcon) {
@@ -489,7 +478,7 @@ Source.prototype = {
         this._isTrayIcon = true;
     },
 
-    _notificationClicked: function(notification) {
+    open: function(notification) {
         this.openApp();
     },
 
@@ -498,6 +487,16 @@ Source.prototype = {
             this.destroy();
     },
 
+    _appStateChanged: function() {
+        // Destroy notification sources when their apps exit.
+        // The app exiting would normally result in a tray icon being removed,
+        // so it should be ok to destroy the source associated with a tray icon
+        // here too, however we just let that happen through the code path
+        // associated with the tray icon being removed.
+        if (!this._isTrayIcon && this.app.get_state() == Shell.AppState.STOPPED)
+            this.destroy();
+    },
+
     openApp: function() {
         if (this.app == null)
             return;
@@ -507,5 +506,13 @@ Source.prototype = {
             let mostRecentWindow = windows[0];
             Main.activateWindow(mostRecentWindow);
         }
+    },
+
+    destroy: function() {
+        if (this.app && this._appStateChangedId) {
+            this.app.disconnect(this._appStateChangedId);
+            this._appStateChangedId = 0;
+        }
+        MessageTray.Source.prototype.destroy.call(this);
     }
 };
